#!/usr/bin/python

# Copyright (c) 2000-2007 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

from xpcom import components, nsError, ServerException

class koAsyncOperationBase(object):
    """Utility class to help creating asynchronous operations"""

    _com_interfaces_ = components.interfaces.koIAsyncOperation

    STATUS_RUNNING = components.interfaces.koIAsyncOperation.STATUS_RUNNING
    STATUS_STOPPING = components.interfaces.koIAsyncOperation.STATUS_STOPPING

    def __init__(self, run_function, *args, **kwargs):
        # status can only be one of: running, stopping
        self.status = self.STATUS_RUNNING
        self.func = run_function
        self.args = args
        self.kwargs = kwargs

    # stop the asynchronous command. Does not have to be implemented,
    # which by default will raise an exception "Cannot be stopped"
    def stop(self):
        raise ServerException(nsError.NS_ERROR_FAILURE, "Cannot be stopped")

    # Private method: called by the koIAsyncService
    def run(self):
        return self.func(*self.args, **self.kwargs)
