/* Copyright (c) 2003-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Komodo's "Breakpoints" tab (in the bottom pane).
 *
 * This tab works with the koIDBGPBreakpointManager service to provide a UI
 * to all relevant breakpoints.
 *
 * Dev Notes:
 * This module initializes itself the first time the "Breakpoints" tab is
 * made visible.
 */

//---- globals

var breakpointsTabLog = ko.logging.getLogger('breakpointsTab');
//breakpointsTabLog.setLevel(ko.logging.LOG_DEBUG);

var _gBreakpointsTab_initialized = false;
var gBreakpointsTabMgr = null;

function BreakpointsTabManager_Initialize() {
    // If have not yet initialized the tab, then do so.
    try {
        if (!_gBreakpointsTab_initialized) {
            var breakpointMgr = Components.classes['@activestate.com/koDBGPBreakpointManager;1']
                            .getService(Components.interfaces.koIDBGPBreakpointManager);
            gBreakpointsTabMgr = new BreakpointsTabManager(breakpointMgr);
            _gBreakpointsTab_initialized = true;
        }
    } catch(ex) {
        breakpointsTabLog.exception(ex);
    }
}

//---- Breakpoints Tab Manager
// One instance of this is created. This implements the
// koIBreakpointsTabManager interface for use by PyXPCOM components.

function BreakpointsTabManager(breakpointMgr) {
    try {
        this.controller = new BreakpointManagerItemsController();
        this._widgets = new Object();
        this._widgets.tab = document.getElementById("breakpoints_tab");
        this._widgets.tabpanel = document.getElementById("breakpoints-tabpanel");
        this._widgets.tree = document.getElementById("breakpoints-tree");
        this._widgets.treeBody = document.getElementById("breakpoints-body");
        this._widgets.treeColumns = {
            "state": document.getElementById("breakpoints-state")
        }
        this._widgets.buttons = {
            "delete": document.getElementById("breakpoints-delete-button"),
            "delete-all": document.getElementById("breakpoints-delete-all-button"),
            "toggle-state": document.getElementById("breakpoints-toggle-state-button"),
            "go-to-source-code": document.getElementById("breakpoints-go-to-source-code-button"),
            "properties": document.getElementById("breakpoints-properties-button")
        }

        this.mgr = breakpointMgr;
        breakpointMgr.tabMgr = this;

        // Bind the nsITreeView instance to the <tree>.
        var boxObject = this._widgets.tree.treeBoxObject
                .QueryInterface(Components.interfaces.nsITreeBoxObject);
        if (boxObject.view == null) {
            //TODO: is showing this truly necessary..., if so, have to move
            //      out to when it is first shown
            // We are in a collapsed state -- we need to force the tree to be
            // visible before we can assign the view to it.
            breakpointsTabLog.debug("manager "+this.id+" initialize(): forcing it to show");
            this.show();
        }
        boxObject.view = this.mgr;

        // Fake an "onselect" so the toolbar gets initialized appropriately.
        this.onSelect();
    } catch(ex) {
        breakpointsTabLog.exception(ex);
    }
}
BreakpointsTabManager.prototype.constructor = BreakpointsTabManager;


BreakpointsTabManager.prototype.QueryInterface = function (iid) {
    if (!iid.equals(Components.interfaces.koIBreakpointsTabManager) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
}


// Number of breakpoints changed: disable/enable the toolbar buttons
// as appropriate.
BreakpointsTabManager.prototype.numBreakpointsChanged = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.numBreakpointsChanged()");
    try {
        // "delete-all" button: enabled when there are 1 or more breakpoints
        // "toggle-state" button: enabled when there are 1 or more breakpoints
        var names = ["delete-all", "toggle-state"];
        var i;
        if (this.mgr.rowCount == 0) {
            for (i = 0; i < names.length; ++i) {
                this._widgets.buttons[names[i]].setAttribute("disabled", "true");
            }
        } else {
            for (i = 0; i < names.length; ++i) {
                this._widgets.buttons[names[i]].removeAttribute("disabled");
            }
        }
    } catch(ex) {
        breakpointsTabLog.exception(ex);
    }
}


// Called when the "Breakpoints" tab is being given the focus.
BreakpointsTabManager.prototype.focus = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.focus()");
    try {
        // Put focus on the tree body.
        this._widgets.treeBody.focus();
    } catch(ex) {
        breakpointsTabLog.exception(ex);
    }
}


BreakpointsTabManager.prototype.show = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.show()");
    try {
        // If necessary, open output pane in komodo.xul
        ko.uilayout.ensureOutputPaneShown(window);

        // switch to proper tab
        var tabsWidget = document.getElementById("output_tabs");
        var tabWidget = document.getElementById(this._idprefix+"_tab");
        tabsWidget.selectedItem = tabWidget;

        // (controversial?) give the find results tab the focus
        //XXX This does not work because the input buffer is passing focus back
        //    to the editor.
        var findResultsWidget = document.getElementById(this._idprefix+"-body");
        findResultsWidget.focus();
    } catch(ex) {
        breakpointsTabLog.exception(ex);
    }
}


BreakpointsTabManager.prototype.onKeyPress = function(event)
{
    breakpointsTabLog.debug("BreakpointsTabManager.onKeyPress(event)");
    //ko.logging.dumpEvent(event);
    try {
        if (event.keyCode == 13 /* <Enter> */) {
            // As in Explorer, any key modifier should result in opening
            // the properties page.
            if (event.altKey || event.ctrlKey || event.shiftKey ||
                event.metaKey) {
                this.breakpointProperties();
            } else {
                this.goToSourceCode();
            }
            event.cancelBubble = true;
            return false;
        } else if (event.isChar && event.charCode == 32 /* space */) {
            var row = this.mgr.selection.currentIndex;
            if (row != -1) {
                var col;

                var boxObject = this._widgets.tree.treeBoxObject
                        .QueryInterface(Components.interfaces.nsITreeBoxObject);
                col = boxObject.columns.getNamedColumn("breakpoints-state");



                this.mgr.cycleCell(row, col);
            }
            event.cancelBubble = true;
            return false;
        } else if (event.keyCode == 46 /* <Del> */) {
            this.deleteBreakpoint();
            event.cancelBubble = true;
            return false;
        }
        //XXX For many keys we get this warning:
        //   WARN: commands: Command cmd_lineNext has no controller.
        // cancel the event bubbling for keys that will be commonly used
        // in this pane so we dont get the warnings.
        if (event.keyCode == event.DOM_VK_UP ||
            event.keyCode == event.DOM_VK_DOWN ||
            event.keyCode == event.DOM_VK_LEFT ||
            event.keyCode == event.DOM_VK_RIGHT) {
            event.cancelBubble = true;
        }
        return true;
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
    return false;
}


BreakpointsTabManager.prototype.onClick = function(event)
{
    breakpointsTabLog.debug("BreakpointsTabManager.onClick(event)");
    try {
        // c.f. mozilla/mailnews/base/resources/content/threadPane.js
        var t = event.originalTarget;

        // unmodified left single-click on a column: sort by that column
        if (event.detail == 1 && event.button == 0 &&
            t.localName == "treecol" &&
            event.shiftKey == false && event.ctrlKey == false &&
            event.altKey == false && event.metaKey == false) {
            this.mgr.sort(t.id);
            this._updateSortIndicators(t.id);
        }

        // unmodified left double-click in the tree body
        else if (event.detail == 2 && event.button == 0 &&
                 t.localName == "treechildren" &&
                 event.shiftKey == false && event.ctrlKey == false &&
                 event.altKey == false && event.metaKey == false) {
            this.goToSourceCode();
        }
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}


// Breakpoints tree selection change: disable/enable the toolbar buttons
// as appropriate.
BreakpointsTabManager.prototype.onSelect = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.onSelect()");
    try {
        window.updateCommands('breakpoint_selection_changed');
    } catch(ex) {
        breakpointsTabLog.exception(ex);
    }
}


// Jump to the location of the currently selected breakpoint.
BreakpointsTabManager.prototype.goToSourceCode = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.goToSourceCode()");
    try {
        var row = this.mgr.selection.currentIndex;
        if (row == -1) {
            return; // no current selection: abort
        }

        // Open and/or switch to the appropriate file.
        var bp = this.mgr.currentBreakpoint;
        var uri = bp.filename;
        var view = null;

        var isRemoteURI = uri.indexOf("dbgp:")==0;
        if (gPrefs.getBooleanPref('debuggerFindLocalFile') && isRemoteURI) {
            var l_uri = uri.slice(8);
            var fileSvc = Components.classes["@activestate.com/koFileService;1"].
                            getService();
            var file = fileSvc.getFileFromURI(l_uri);
            if (file.exists) {
                uri = l_uri;
                isRemoteURI = false;
            }
        }
        if (isRemoteURI && !DBG.currentSession) {
            ko.dialogs.alert("Unable to open remote debugger file.");
            return;
        }

        try {
            //XXX Would like to have an optional argument to ko.open.URI()
            //    to NOT offer to create a file that is not found. In this
            //    use-case this should just report an error.
            view = ko.open.URI(uri);
        } catch(ex) {
            var displayPath = ko.uriparse.displayPath(uri);
            var msg = "Error opening '"+displayPath+"'.";
            ko.dialogs.alert(msg);
            breakpointsTabLog.exception(ex, msg);
            return;
        }
        if (view) {
            // Jump to the proper line.
            var line = bp.lineno;
            if (line >= 1) {
                view.scimoz.gotoLine(Math.max(line - 1, 0)); // scimoz is 0-indexed
            }
            view.scimoz.chooseCaretX();

            // Transfer focus to the editor.
            view.setFocus();
        }
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}


BreakpointsTabManager.prototype.newBreakpoint = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.newBreakpoint()");
    try {
        Breakpoints_NewBreakpoint();
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}


BreakpointsTabManager.prototype.newSpawnpoint = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.newSpawnpoint()");
    try {
        Breakpoints_NewSpawnpoint();
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}


BreakpointsTabManager.prototype.deleteBreakpoint = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.deleteBreakpoint()");
    try {
        var row = this.mgr.selection.currentIndex;
        if (row == -1) {
            return; // no current selection: abort
        }

        // Ensure the user wants to delete this breakpoint.
        var question = "Are you sure you want to delete the '"+
                       this.mgr.currentBreakpointName+"' breakpoint?";
        var answer = ko.dialogs.okCancel(question, "OK", null,
                                     "Confirm Breakpoint Delete",
                                     "confirm_delete_breakpoint");
        if (answer == "OK") {
            this.mgr.removeBreakpoint(this.mgr.currentBreakpointGuid);

            // Select a reasonable row after deletion.
            if (row > this.mgr.rowCount-1) {
                row = this.mgr.rowCount-1;
            }
            if (row >= 0) {
                this.mgr.selection.select(row);
            } else {
                this.mgr.selection.clearSelection();
            }
        }
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}


BreakpointsTabManager.prototype.deleteAllBreakpoints = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.deleteAllBreakpoints()");
    try {
        var didit = Breakpoints_DeleteAllBreakpoints();
        if (didit) {
            this.mgr.selection.clearSelection();
        }
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}

BreakpointsTabManager.prototype.toggleBreakpointState = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.toggleBreakpointState()");
    try {
        this.mgr.toggleBreakpointState(this.mgr.currentBreakpointGuid);
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}

BreakpointsTabManager.prototype.toggleAllBreakpointStates = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.toggleAllBreakpointStates()");
    try {
        this.mgr.toggleAllBreakpointStates();
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}


BreakpointsTabManager.prototype.breakpointProperties = function()
{
    breakpointsTabLog.debug("BreakpointsTabManager.breakpointProperties()");
    try {
        var row = this.mgr.selection.currentIndex;
        if (row == -1) {
            return; // no current selection: abort
        }

        Breakpoints_BreakpointProperties(this.mgr.currentBreakpoint);
    } catch (ex) {
        breakpointsTabLog.exception(ex);
    }
}


BreakpointsTabManager.prototype._updateSortIndicators = function(sortId)
{
    return; //XXX disable this for now (need to solve so other problems first)
    var sortedColumn = null;

    // set the sort indicator on the column we are sorted by
    if (sortId) {
        sortedColumn = document.getElementById(sortId);
        if (sortedColumn) {
            var sortDirection = sortedColumn.getAttribute("sortDirection");
            if (sortDirection && sortDirection == "ascending") {
                sortedColumn.setAttribute("sortDirection", "descending");
            } else {
                sortedColumn.setAttribute("sortDirection", "ascending");
            }
        }
    }

    // remove the sort indicator from all the columns
    // except the one we are sorted by
    var currCol = document.getElementById("breakpoints-tree").firstChild.firstChild;
    while (currCol) {
        while (currCol && currCol.localName != "treecol") {
            currCol = currCol.nextSibling;
        }
        if (currCol && (currCol != sortedColumn)) {
            currCol.removeAttribute("sortDirection");
        }
        if (currCol) {
            currCol = currCol.nextSibling;
        }
    }
}


function BreakpointManagerItemsController() {
    window.controllers.appendController(this);
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
BreakpointManagerItemsController.prototype = new Controller();
BreakpointManagerItemsController.prototype.constructor =BreakpointManagerItemsController;

BreakpointManagerItemsController.prototype.destroy = function() {
    window.controllers.removeController(this);
}

BreakpointManagerItemsController.prototype.is_cmd_breakpoint_add_enabled = function () {
    return true;
}

BreakpointManagerItemsController.prototype.do_cmd_breakpoint_add = function() {
    return gBreakpointsTabMgr.newBreakpoint();
}

BreakpointManagerItemsController.prototype.is_cmd_spawnpoint_add_enabled = function () {
    return true;
}

BreakpointManagerItemsController.prototype.do_cmd_spawnpoint_add = function() {
    return gBreakpointsTabMgr.newSpawnpoint();
}

BreakpointManagerItemsController.prototype.is_cmd_breakpoint_goto_enabled = function () {
    return gBreakpointsTabMgr && gBreakpointsTabMgr.mgr.selection.currentIndex >= 0;
}

BreakpointManagerItemsController.prototype.do_cmd_breakpoint_goto = function() {
    gBreakpointsTabMgr.goToSourceCode();
    return true;
}

BreakpointManagerItemsController.prototype.is_cmd_breakpoint_toggle_enabled = function () {
    return gBreakpointsTabMgr && gBreakpointsTabMgr.mgr.selection.currentIndex >= 0;
}

BreakpointManagerItemsController.prototype.do_cmd_breakpoint_toggle = function() {
    gBreakpointsTabMgr.toggleBreakpointState();
    return true;
}

BreakpointManagerItemsController.prototype.is_cmd_breakpoint_toggle_all_enabled = function () {
    return gBreakpointsTabMgr && gBreakpointsTabMgr.mgr.rowCount > 0;
}

BreakpointManagerItemsController.prototype.do_cmd_breakpoint_toggle_all = function() {
    gBreakpointsTabMgr.toggleAllBreakpointStates();
    return true;
}

BreakpointManagerItemsController.prototype.is_cmd_breakpoint_delete_enabled = function () {
    return gBreakpointsTabMgr && gBreakpointsTabMgr.mgr.selection.currentIndex >= 0;
}

BreakpointManagerItemsController.prototype.do_cmd_breakpoint_delete = function() {
    gBreakpointsTabMgr.deleteBreakpoint();
    return true;
}

BreakpointManagerItemsController.prototype.is_cmd_breakpoint_delete_all_enabled = function () {
    return gBreakpointsTabMgr && gBreakpointsTabMgr.mgr.rowCount > 0;
}

BreakpointManagerItemsController.prototype.do_cmd_breakpoint_delete_all = function() {
    gBreakpointsTabMgr.deleteAllBreakpoints();
    return true;
}

BreakpointManagerItemsController.prototype.is_cmd_breakpoint_properties_enabled = function () {
    return gBreakpointsTabMgr && gBreakpointsTabMgr.mgr.selection.currentIndex >= 0;
}

BreakpointManagerItemsController.prototype.do_cmd_breakpoint_properties = function() {
    window.setTimeout('gBreakpointsTabMgr.breakpointProperties();', 1);
    return true;
}


//---- interface routines

// Called when the Breakpoints tab is being selected/made visible.
//
// The first time this is called, the tab initializes itself.
function BreakpointsTab_Select()
{
    breakpointsTabLog.debug("BreakpointsTab_Select()");
    try {
        gBreakpointsTabMgr.focus();
    } catch(ex) {
        breakpointsTabLog.exception(ex);
    }
}


